
//
// XML storage classes version 1.3
//
// XMLDocReader.java
//
// Copyright (c) 2004, 2005, 2007, 2008 Martin Fuchs <martin-fuchs@gmx.net>
//


/*

 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in
	the documentation and/or other materials provided with the
	distribution.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

*/

package de.mfuchs.xmlstorage;

import java.io.IOException;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.ext.DefaultHandler2;


/**
 * XMLDocReader is the XML file reader of XMLStorage.
 */
final class XMLDocReader extends DefaultHandler2
{
	XMLDocReader(XMLNode node) throws ParserConfigurationException, SAXException
	{
		_pos = new XMLPos(node);
		_lastTag = TAG_NONE;

		SAXParserFactory factory = SAXParserFactory.newInstance();
		_parser = factory.newSAXParser();

		 // activate CDATA notifications
		_parser.setProperty("http://xml.org/sax/properties/lexical-handler", this);
	}

	 /// read XML stream into XML tree below _pos
	void read(InputSource source) throws SAXException, IOException
	{
		 // read XML data
		_parser.parse(source, this);

		 // process last data pieces
		if (_pos.cur()._children.isEmpty())
			_pos.cur()._trailing += _content;
		else
			_pos.cur()._children.getLast()._trailing += _content;

		_content = "";
		_isContentCDATA = false;
	}

	private XMLPos	_pos;
	private SAXParser _parser;

	private String	_content = "";
	private StringBuffer _instructions = new StringBuffer(); 
 
	private static final int TAG_NONE = 0;
	private static final int TAG_START = 1;
	private static final int TAG_END = 2;
	private int _lastTag;	// enum


	public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
	{
		 // search for content end leaving only whitespaces for _end_leading
		int i = _content.length();
		for (; i>0; i--) {
			char ch = _content.charAt(i-1);
			if (!Character.isWhitespace(ch))
				break;
		}

		if (i > 0)
			if (_pos.cur()._children.isEmpty()) {	// no children in last node?
				if (_lastTag == TAG_START)
					_pos.cur()._content += _content.substring(0, i);
				else if (_lastTag == TAG_END)
					_pos.cur()._trailing += _content.substring(0, i);
				// else TAG_NONE -> don't store white space in root node
			} else
				_pos.cur()._children.getLast()._trailing += _content.substring(0, i);

		String leading = "";

		if (i < _content.length())
			leading = _content.substring(i);

		XMLNode node = new XMLNode(qName, leading);
		//XMLNode node = new XMLNode(XMLHelper.String_from_XML_Char(qName), leading);

		_pos.addDown(node);

		for(int n=0; n<attributes.getLength(); ++n) {
			final String attrName = attributes.getQName(n);
			final String attrValue = attributes.getValue(n);

			node._attributes.put(attrName, attrValue);
			//node.put(XMLHelper.String_from_XML_Char(attrName), XMLHelper.String_from_XML_Char(attr_value));
		}

		_lastTag = TAG_START;
		_content = "";
		_isContentCDATA = false;
	}

	public void endElement(String uri, String localName, String qName) throws SAXException
	{
		int i;

		if (_isContentCDATA)
			i = _content.length();
		else {
			 // search for content end leaving only whitespaces for leading
			i = _content.length();
			for (; i>0; i--) {
				char ch = _content.charAt(i-1);
				if (!Character.isWhitespace(ch))
					break;
			}
		}

		if (i > 0)
			if (_pos.cur()._children.isEmpty()) // no children in current node?
				_pos.cur()._content += _content.substring(0, i);
			else
				if (_lastTag == TAG_START)
					_pos.cur()._content += _content.substring(0, i);
				else
					_pos.cur()._children.getLast()._trailing += _content.substring(0, i);

		if (i < _content.length())
			_pos.cur()._endLeading = _content.substring(i);

		_pos.back();

		_lastTag = TAG_END;
		_content = "";
		_isContentCDATA = false;
	}

	public void characters(char[] ch, int start, int length) throws SAXException
	{
		 // encode _content compatible to XMLNode.setContent()
		_content += XMLHelper.encodeXMLString(new String(ch, start, length));

		if (_inCDATASection)
			_isContentCDATA = true;
	}


	 // CDATA-Handling

	private boolean _inCDATASection = false;
	private boolean _isContentCDATA = false;

	public void startCDATA() throws SAXException
	{
		_inCDATASection = true;
	}

	public void endCDATA() throws SAXException
	{
		_inCDATASection = false;
	}


	public void processingInstruction(String target, String data)
		throws SAXException
	{
		_instructions.append("<?").append(target)
			.append(" ").append(data).append("?>");
	}

	public String getInstructions()
	{
		return _instructions.toString();
	}
}
