
//
// XML storage classes version 1.3
//
// XMLNode.java
//
// Copyright (c) 2004, 2005, 2007, 2008 Martin Fuchs <martin-fuchs@gmx.net>
//


/*

 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in
	the documentation and/or other materials provided with the
	distribution.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

*/

package de.mfuchs.xmlstorage;

import java.io.IOException;
import java.io.Writer;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.TreeMap;


/**
 * XMLNode is the in memory representation of an XML node
 * in the XMLStorage library.
 */
public class XMLNode extends AbstractXMLNode
{
	public XMLNode(final String name)
	{
	    _name = name;
	}

	public XMLNode(final String name, final String leading)
	{
	     _name = name;
	     _leading = leading;
	}

	public XMLNode(final XMLNode other)
	{
	    assign(other);
	}

    protected String _name;

	public String getName()
	{
		return _name;
	}

    public String toString()
    {
        return _name;
    }

    /**
     * Children contains the list of immediate child nodes.
     */
	public static class Children extends LinkedList<XMLNode>
	{
        private static final long serialVersionUID = 1L;

        public void assign(final Children other)
        {
    		this.clear();

    		for(XMLNode node: other)
    			addLast(new XMLNode(node));
        }

        public void clear()
        {
    		while(!isEmpty()) {
    			XMLNode node = getLast();
    			removeLast();

    			node.clear();
    		}
        }
	}

	/**
	 * AttributeMap is a sorted map of attributes.
	 */
	public static class AttributeMap extends TreeMap<String, String>
	{
        private static final long serialVersionUID = 1L;
	}

	protected Children _children = new Children();
	protected AttributeMap _attributes = new AttributeMap();

	protected String	_leading = "";
	protected String	_content = "";
	protected String	_endLeading = "";
	protected String	_trailing = "";

	public void clear()
	{
		_leading = "";
		_content = "";
		_endLeading = "";
		_trailing = "";

		_attributes.clear();
		_children.clear();

		_name = "";
	}

	public void assign(final XMLNode other)
	{
	    _children.assign(other._children);

	 	_attributes = (AttributeMap) other._attributes.clone();

		_leading = other._leading;
		_content = other._content;
		_endLeading = other._endLeading;
		_trailing = other._trailing;
	}

	 /// add a new child node
	public void addChild(XMLNode child)
	{
		_children.addLast(child);
	}

	 /// read only access to an attribute
	public String get(final String attrName)
	{
        String found = _attributes.get(attrName);

		if (found != null)
			return found;
		else
			return "";
	}

     /// read only access to an attribute with default value
    public String get(final String attrName, final String def)
    {
        String found = _attributes.get(attrName);

        if (found != null)
            return found;
        else
            return def;
    }

	 /// write access to an attribute
	public void put(final String attrName, final String value)
	{
		_attributes.put(attrName, value);
	}

	 /// convenient value access in children node
	public String getSubattrib(final String nodeName, final String attrName)
	{
		final AbstractXMLNode node = find(nodeName);

		if (node != null)
			return node.get(attrName);
		else
			return "";
	}

	public String getSubattrib(final String nodeName, final String attrName, int n)
	{
		final AbstractXMLNode node = find(nodeName, n);

		if (node != null)
			return node.get(attrName);
		else
			return "";
	}

	/// convenient storage of distinct values in children node
	public void putSubattrib(final String nodeName, final String attrName, final String value)
	{
		XMLNode node = find(nodeName);

		if (node == null) {
			node = new XMLNode(nodeName);
			addChild(node);
		}

		node.put(attrName, value);
	}

	public void putSubattrib(final String nodeName, final String attrName, final String value, int n)
	{
		XMLNode node = find(nodeName, n);

		if (node == null) {
			node = new XMLNode(nodeName);
			addChild(node);
		}

		node.put(attrName, value);
	}

	public Children getChildren()
	{
		return _children;
	}

	public String getContent()
	{
		return XMLHelper.decodeXMLString(_content);
	}

	public void setContent(final String s)
	{
        setContent(s, false);
	}

    public void setContent(final String s, boolean cdata)
    {
        _content = XMLHelper.encodeXMLString(s, cdata);
    }

	 // enum WRITE_MODE
	public static final int FORMAT_SMART	= 0;	/// preserve original white space and comments if present; pretty print otherwise
	public static final int FORMAT_ORIGINAL	= 1;	/// write XML stream preserving original white space and comments
	public static final int FORMAT_PRETTY	= 2; 	/// pretty print node to stream without preserving original white space

	public void write(Writer out) throws Exception
	{
	    write(out, FORMAT_SMART, 0);
	}

	public void write(Writer out, int mode) throws Exception
	{
	    write(out, mode, 0);
	}

	 /// write node with children tree
	public void write(Writer out, int mode/*=FORMAT_SMART*/, int indent/*=0*/) throws IOException
	{
		switch(mode) {
		  case FORMAT_PRETTY:
			prettyWriteWorker(out, indent);
			break;

		  case FORMAT_ORIGINAL:
			writeWorker(out, indent);
			break;

		default:	 // FORMAT_SMART
			smartWriteWorker(out, indent);
		}
	}

	protected XMLNode getFirstChild()
	{
		if (!_children.isEmpty())
			return _children.getFirst();
		else
			return null;
	}

	/**
	 * Find a children node by name.
	 * @param nodeName
	 * @return
	 */
	public XMLNode find(final String nodeName)
	{
		for(XMLNode child: _children)
		    if (child._name.equals(nodeName))
				return child;

		return null;
	}

	/**
	 * Find the n'th children node with tag nodeName.
	 * @param nodeName
	 * @param n
	 * @return
	 */
	public XMLNode find(final String nodeName, int n)
	{
		for(XMLNode child: _children)
		    if (child._name.equals(nodeName))
		        if (n-- == 0)
		            return child;

		return null;
	}

	 /**
	  * Search matching child node identified by key name and an attribute value.
	  * @param nodeName
	  * @param attrName
	  * @param attrValue
	  * @return XMLNode
	  */
	public XMLNode find(final String nodeName, final String attrName, final String attrValue)
	{
		for(XMLNode child: _children)
			if (child._name.equals(nodeName) && child.get(attrName).equals(attrValue))
				return child;

		return null;
	}

	 /**
	  * Search matching child node identified by key name and an attribute value.
	  * @param nodeName
	  * @param attrName
	  * @param attrValue
	  * @param n
	  * @return XMLNode
	  */
	public XMLNode find(final String nodeName, final String attrName, final String attrValue, int n)
	{
		for(XMLNode node: _children)
			if (node._name.equals(nodeName) && node.get(attrName).equals(attrValue))
			    if (n-- == 0)
			        return node;

		return null;
	}

     /// XPath find function (const function)
    XMLNode findRelative(String path)
    {
        XMLNode node = this;

         // parse relative path
        while(path.length() > 0) {
            int idxSlash = path.indexOf('/');
            if (idxSlash == 0)
                return null;

            int l = idxSlash!=-1? idxSlash: path.length();
            String comp = path.substring(0, l);
            path = path.substring(l);

             // look for [n] and [@attrName="attrValue"] expressions in path components
            int idxBracket = comp.indexOf('[');
            l = idxBracket!=-1? idxBracket: comp.length();
            String childName = comp.substring(0, l);
            String attrName="", attrValue="";

            int n = 0;
            if (idxBracket != -1) {
                String expr = XMLHelper.unescape(comp.substring(idxBracket), '[', ']');

                try {
                    n = Integer.parseInt(expr);    // read index number
                } catch(NumberFormatException e) {
                    n = 0;
                }

                if (n != 0)
                    n = n - 1;  // convert into zero based index

                int idxAt = expr.indexOf('@');

                if (idxAt != -1) {
                    expr = expr.substring(idxAt+1);
                    int idxEqual = expr.indexOf('=');

                     // read attribute name and value
                    if (idxEqual != -1) {
                        attrName = XMLHelper.unescape(expr.substring(0, idxEqual));
                        attrValue = XMLHelper.unescape(expr.substring(idxEqual+1));
                    }
                }
            }

            XMLNode child;

            if (attrName.length() == 0)
                 // search n.th child node with specified name
                child = node.find(childName, n);
            else
                 // search n.th child node with specified name and matching attribute value
                child = node.find(childName, attrName, attrValue, n);

            if (child == null)
                return null;

            node = child;

            if (path.length()>0 && path.charAt(0)=='/')
                path = path.substring(1);
        }

        return node;
    }

     /// relative XPath create function
    AbstractXMLNode createRelative(String path)
    {
        XMLNode node = this;

         // parse relative path
        while(path.length() > 0) {
            int idxSlash = path.indexOf('/');
            if (idxSlash == 0)
                return null;

            int l = idxSlash!=-1? idxSlash: path.length();
            String comp = path.substring(0, l);
            path = path.substring(l);

             // look for [n] and [@attrName="attrValue"] expressions in path components
            int idxBracket = comp.indexOf('[');
            l = idxBracket!=-1? idxBracket: comp.length();
            String childName = comp.substring(0, l);
            String attrName="", attrValue="";

            int n = 0;
            if (idxBracket != -1) {
                String expr = XMLHelper.unescape(comp.substring(idxBracket), '[', ']');

                try {
                    n = Integer.parseInt(expr);    // read index number
                } catch(NumberFormatException e) {
                    n = 0;
                }

                if (n != 0)
                    n = n - 1;  // convert into zero based index

                int idxAt = expr.indexOf('@');

                if (idxAt != -1) {
                    expr = expr.substring(idxAt+1);
                    int idxEqual = expr.indexOf('=');

                     // read attribute name and value
                    if (idxEqual != -1) {
                        attrName = XMLHelper.unescape(expr.substring(0, idxEqual));
                        attrValue = XMLHelper.unescape(expr.substring(1));
                    }
                }
            }

            XMLNode child;

            if (attrName.length() == 0)
                 // search n.th child node with specified name
                child = node.find(childName, n);
            else
                 // search n.th child node with specified name and matching attribute value
                child = node.find(childName, attrName, attrValue, n);

            if (child == null) {
                child = new XMLNode(childName);
                node.addChild(child);

                if (attrName.length() > 0)
                    node.put(attrName, attrValue);
            }

            node = child;

            if (path.length()>0 && path.charAt(0)=='/')
                path = path.substring(1);
        }

        return node;
    }

/*@@ TODO
	XMLNode getChildRelative(inout char* ptr, bool create)
	{
	}
*/

	 /**
	  * Write node with children tree using original white space.
	  * @param out
	  * @param indent
	  * @throws IOException
	  */
	protected void writeWorker(Writer out, int indent) throws IOException
	{
		out.write(_leading);
		out.write('<');
		out.write(XMLHelper.encodeXMLString(_name));

		for(Map.Entry<String,String> it: _attributes.entrySet()) {
            String attrValue = it.getValue();

            if (attrValue != null) {
    			out.write(' ');
    			out.write(XMLHelper.encodeXMLString(it.getKey()));
    			out.write("=\"");
    			out.write(XMLHelper.encodeXMLString(attrValue));
    			out.write("\"");
            }
		}


		if (!_children.isEmpty() || _content.length()>0) {
		    out.write('>');
		    out.write(_content);

			for(XMLNode child: _children)
				child.writeWorker(out, indent+1);

			out.write(_endLeading);
			out.write("</");
			out.write(XMLHelper.encodeXMLString(_name));
			out.write('>');
		} else
		    out.write("/>");

		out.write(_trailing);
	}

	public static final String XML_INDENT_SPACE = "  ";

	protected void prettyWriteWorker(Writer out, int indent) throws IOException
	{
		for(int i=indent; i-->0; )
			out.write(XML_INDENT_SPACE);

		out.write('<');
		out.write(XMLHelper.encodeXMLString(_name));

		for(Map.Entry<String,String> it: _attributes.entrySet()) {
            String attrValue = it.getValue();

            if (attrValue != null) {
    			out.write(' ');
    			out.write(XMLHelper.encodeXMLString(it.getKey()));
    			out.write("=\"");
    			out.write(XMLHelper.encodeXMLString(attrValue));
    			out.write("\"");
            }
		}

		if (!_children.isEmpty() || _content.length()>0) {
			out.write(">\n");

			for(XMLNode child: _children)
				child.prettyWriteWorker(out, indent+1);

			for(int i=indent; i-->0; )
				out.write(XML_INDENT_SPACE);

			out.write("</");
			out.write(XMLHelper.encodeXMLString(_name));
			out.write(">\n");
		} else
			out.write("/>\n");
	}

	protected void smartWriteWorker(Writer out, int indent) throws IOException
	{
		if (_leading.length() == 0)
			for(int i=indent; i-->0; )
				out.write(XML_INDENT_SPACE);
		else
			out.write(_leading);

		out.write('<');
		out.write(XMLHelper.encodeXMLString(_name));

		for(Map.Entry<String,String> it: _attributes.entrySet()) {
            String attrValue = it.getValue();

            if (attrValue != null) {
    			out.write(' ');
    			out.write(XMLHelper.encodeXMLString(it.getKey()));
    			out.write("=\"");
    			out.write(XMLHelper.encodeXMLString(attrValue));
    			out.write("\"");
            }
		}

		if (_children.isEmpty() && _content.length()==0)
			out.write("/>");
		else {
			out.write('>');

			if (_content.length() == 0)
				out.write('\n');
			else
				out.write(_content);

			Iterator<XMLNode> it = _children.iterator();

			if (it.hasNext()) {
				while(it.hasNext())
					it.next().smartWriteWorker(out, indent+1);

				if (_endLeading.length() == 0)
					for(int i=indent; i-->0; )
						out.write(XML_INDENT_SPACE);
				else
					out.write(_endLeading);
			} else
				out.write(_endLeading);

			out.write("</");
			out.write(XMLHelper.encodeXMLString(_name));
			out.write('>');
		}

		if (_trailing.length() == 0)
			out.write('\n');
		else
			out.write(_trailing);
	}
}
