
//
// XML storage classes version 1.2
//
// XMLPos.java
//
// Copyright (c) 2004, 2005, 2007 Martin Fuchs <martin-fuchs@gmx.net>
//


/*

 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in
	the documentation and/or other materials provided with the
	distribution.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

*/

package de.mfuchs.xmlstorage;

import java.util.Stack;


/**
 * XMLPos is the iterator for XML trees in XMLStorage.
 * @author Martin Fuchs
 */
public class XMLPos
{
    public XMLPos(XMLNode root)
	{
		_root = root;
		_cur = root;
	}

	public XMLPos(final XMLPos other)
	{
	    _root = other._root;
		_cur = other._cur;
	    // don't copy _stack
	}

	/**
	 * Create XML tree iterator positioned at subnode node_name of current other position.
	 * @param other XMLPos
	 * @param nodeName
	 */
	public XMLPos(final XMLPos other, final String nodeName)
	{
	    _root = other._root;
		_cur = other._cur;
	    // don't copy _stack
		smartCreate(nodeName);
	}

	/**
	 * Create XML tree iterator positioned at subnode node_name
	 * with attribute attrName and attribute value attrValue.
	 * @see XMLPos.smart_create
	 * @param other
	 * @param nodeName
	 * @param attrName
	 * @param attrValue
	 */
	public XMLPos(final XMLPos other, final String nodeName, final String attrName, final String attrValue)
	{
	    _root = other._root;
		_cur = other._cur;
	    // don't copy _stack
		smartCreate(nodeName, attrName, attrValue);
	}

	 /// access to current node
	public XMLNode cur()
	{
	    return _cur;
	}

	/**
	 * Return the value of an attribute in the current node or "".
	 * @param attrName
	 * @return
	 */
	public String get(final String attrName)
	{
	    return _cur.get(attrName);
	}

    /**
     * Return the value of an attribute in the current node or a default value.
     * @param attrName
     * @param default value
     * @return
     */
    public String get(final String attrName, final String def)
    {
        return _cur.get(attrName, def);
    }

	/**
	 * Set an attribut value.
	 * @param attrName
	 * @param value
	 */
	public void put(final String attrName, final String value)
	{
	    _cur.put(attrName, value);
	}

	/**
	 * Insert children when building tree.
	 * @param child
	 */
	public void addDown(XMLNode child)
	{
		_cur.addChild(child);
		goTo(child);
	}

	/**
	 * Go back to previous position.
	 * @return true if successfull
	 */
	public boolean back()
	{
		if (!_stack.empty()) {
			_cur = _stack.pop();
			return true;
		} else
			return false;
	}

	/**
	 * Go down to first child.
	 * @return true if successfull
	 */
	public boolean goDown()
	{
		XMLNode node = _cur.getFirstChild();

		if (node != null) {
			goTo(node);
			return true;
		} else
			return false;
	}

	/**
	 * Search for child and go down.
	 * @param name
	 * @return
	 */
	public boolean goDown(final String name)
	{
		XMLNode node = _cur.find(name);

		if (node != null) {
			goTo(node);
			return true;
		} else
			return false;
	}

	 /// move X-Path like to position in XML tree
	public boolean go(String path)
    {
        XMLNode node = _cur;

         // Is this an absolute path?
        if (path.charAt(0) == '/') {
            node = _root;
            path = path.substring(1);
        }

        node = node.findRelative(path);

        if (node != null) {
            goTo(node);
            return true;
        } else
            return false;
    }

	/**
	 * Create node and move to it.
	 * @param name
	 */
	public void create(final String name)
	{
		addDown(new XMLNode(name));
	}

	/**
	 * Create node if not already existing and move to it.
	 * @param name
	 */
	public void smartCreate(final String name)
	{
		XMLNode node = _cur.find(name);

		if (node != null)
			goTo(node);
		else
			addDown(new XMLNode(name));
	}

	/**
	 * Search matching child node identified by key name and an attribute value.
	 * @param name
	 * @param attrName
	 * @param attrValue
	 */
	public void smartCreate(final String name, final String attrName, final String attrValue)
	{
		XMLNode node = _cur.find(name, attrName, attrValue);

		if (node != null)
			goTo(node);
		else {
			node = new XMLNode(name);
			addDown(node);
			node.put(attrName, attrValue);
		}
	}

	/**
	 * @return content string
	 */
	public String getContent()
	{
	    return _cur.getContent();
	}

	/**
	 * @param content string
	 */
	public void setContent(final String s)
	{
	    setContent(s, false);
	}

    /**
     * @param content string
     */
    public void setContent(final String s, boolean cdata)
    {
        _cur.setContent(s, cdata);
    }

	public String toString()
	{
		return _cur.toString();
	}

	protected XMLNode	_root;
	protected XMLNode	_cur;
	protected Stack<XMLNode> _stack = new Stack<XMLNode>();

	/**
	 * Go to the specified node.
	 * @param child
	 */
	protected void goTo(XMLNode child)
	{
		_stack.push(_cur);
		_cur = child;
	}

	/**
	 * Set an attribute in a subnode.
	 */
	public void putSubattrib(final String nodeName, final String attrName, final String value)
	{
		_cur.putSubattrib(nodeName, attrName, value);
	}
}
