
//
// XML storage classes version 1.3
//
// XMLStreamReader.java
//
// Copyright (c) 2008 Martin Fuchs <martin-fuchs@gmx.net>
//
// XMLStreamReader requires Java 1.6
//


/*

 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in
	the documentation and/or other materials provided with the
	distribution.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

*/

package de.mfuchs.xmlstorage.stream;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.Iterator;

import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.events.XMLEvent;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;

import de.mfuchs.xmlstorage.XMLHelper;


/**
 * XMLStreamReader is the XML file reader of XMLStorage.
 * XMLStreamReader requires Java 1.6.
 */
public class XMLStreamReader implements Iterable<XMLStreamNode>
{
    private InputStream _istream;
	protected javax.xml.stream.XMLStreamReader _reader;

	protected XMLStreamNode _node;

	protected String		_content = "";
    protected StringBuffer	_instructions = new StringBuffer(); 

	int _level	= 0;

	public XMLStreamReader(String path) throws FileNotFoundException, XMLStreamException
	{
	    _istream = new FileInputStream(path);

	    StreamSource source = new StreamSource(_istream);
	    source.setSystemId(path);

	    init(source);
	}

	public XMLStreamReader(Source source) throws XMLStreamException
	{
		init(source);
	}

	private void init(Source source) throws XMLStreamException
	{
		_node = new XMLStreamNode(this);

		XMLInputFactory factory = XMLInputFactory.newInstance();

		_reader = factory.createXMLStreamReader(source);
	}

	public void close() throws XMLStreamException, IOException
	{
		_reader.close();
		_reader = null;

		if (_istream != null) {
			_istream.close();
			_istream = null;
		}
	}


	public XMLStreamNode getNode()
	{
		return _node;
	}

	public Object getNodeName()
    {
	    return _node.getName();
    }

	public boolean isAtNode(String nodeName)
    {
	    return _node.getName().equals(nodeName);
    }


    /**
     * XMLStreamIterator iterates over the immediate children of the current XML node.
     */
	private static class XMLStreamIterator implements Iterator<XMLStreamNode>
	{
		XMLStreamIterator(XMLStreamReader reader)
		{
			_reader = reader;

			moveToChild();
		}

        public boolean hasNext()
        {
	        return _hasNext;
        }

        public XMLStreamNode next()
        {
	        moveToNext();

	        return _reader.getNode();
        }

        public void remove()
        {
	        throw new UnsupportedOperationException("remove() not implemented for XMLStreamIterator");
        }

		private final XMLStreamReader _reader;
		private boolean _hasNext;

		private void moveToChild()
		{
	        try {
				_hasNext = _reader.firstChild();
            } catch(XMLStreamException e) {
				_hasNext = false;
	            e.printStackTrace();
            }
		}

		private void moveToNext()
		{
	        try {
				_hasNext = _reader.next();
            } catch(XMLStreamException e) {
				_hasNext = false;
	            e.printStackTrace();
            }
		}
	}

    public Iterator<XMLStreamNode> iterator()
    {
        return new XMLStreamIterator(this);
    }


	/**
	 * Return the value of an attribute in the current node or "".
	 * @param attrName
	 * @return
	 */
	public String get(final String attrName)
	{
        return _node.get(attrName);
	}

    /**
     * Return the value of an attribute in the current node or a default value.
     * @param attrName
     * @param default value
     * @return
     */
    public String get(final String attrName, final String def)
    {
        return _node.get(attrName, def);
    }

	/**
	 * Go down to first child.
	 * @return true if successfull
	 * @throws XMLStreamException 
	 */
	public boolean goDown() throws XMLStreamException
	{
		return firstChild(null);
	}

	/**
	 * Search for child and go down.
	 * @param name
	 * @return
	 * @throws XMLStreamException 
	 */
	public boolean goDown(final String name) throws XMLStreamException
	{
		return firstChild(name);
	}

	 /// move X-Path like to position in XML tree
	public boolean go(String path) throws XMLStreamException
    {
         // Absolute paths are not supported.
        if (path.charAt(0) == '/')
            return false;

        return findRelative(path);
    }

	/**
	 * Move back to the next higher nesting level.
	 * Remark: This implementation does _not_ jump more than one level up like XMLPos.node() when using go() to move down more than one level at a time.
	 * @return true if successfull
	 * @throws XMLStreamException 
	 */
	public boolean back() throws XMLStreamException
    {
		return findNode(null, -1, _level+1, false);
    }

	/**
	 * @return content string
	 */
	public String getContent()
	{
	    return _node.getContent();
	}

	public String toString()
	{
		return _node.toString();
	}


	private void startElement(String nodeName)
	{
		_node.assign(nodeName);
		_node.setAttributes(_reader);

		_content = "";

		++_level;
	}

	private void endElement(String nodeName)
	{
		_node.assign(nodeName);

		_content = "";

		--_level;
	}

	private void characters()
	{
		_content += XMLHelper.encodeXMLString(_reader.getText());
	}


	private void processingInstruction()
	{
		_instructions.append("<?").append(_reader.getPITarget())
			.append(" ").append(_reader.getPIData()).append("?>");
	}

	public String getInstructions()
	{
		return _instructions.toString();
	}


	/**
	 * Move to the first child node.
	 * @return true on success
	 * @throws XMLStreamException
	 */
	public boolean firstChild() throws XMLStreamException
	{
		return findNode(null, _level+1, _level, false);
	}

	/**
	 * Move to the first child node with the specified name.
	 * @return true on success
	 * @throws XMLStreamException
	 */
	public boolean firstChild(String nodeName) throws XMLStreamException
	{
		return findNode(nodeName, _level+1, _level, false);
	}


	/**
	 * Move to the next sibling node independent from its name.
	 * @return true on success
	 * @throws XMLStreamException
	 */
	public boolean next() throws XMLStreamException
	{
		return findNode(null, _level, _level-1, false);
	}

	/**
	 * Move to the next sibling node with the specified name.
	 * @return true on success
	 * @throws XMLStreamException
	 */
	public boolean nextSkipOthers(String nodeName) throws XMLStreamException
	{
		return findNode(nodeName, _level, _level-1, false);
	}

	/**
	 * Move to the next sibling node with the specified name.
	 * @return true on success
	 * @throws XMLStreamException
	 */
	public boolean nextStopOnOthers(String nodeName) throws XMLStreamException
	{
		return findNode(nodeName, _level, _level-1, true);
	}


	private boolean findNode(String targetTag, int targetLevel, int minLevel, boolean stopOnOthers) throws XMLStreamException
	{
		boolean found = false;

		parsingLoop: for(; _reader.hasNext(); _reader.next()) {
			switch(_reader.getEventType()) {
				case XMLEvent.START_ELEMENT: {
					 // stop parsing if we found the target node and its content is already processed.
					if (found)
						break parsingLoop;

					String nodeName = _reader.getLocalName();

					 // check for match in target level and eventually node name
					if (_level+1 == targetLevel) {
						if (targetTag==null || nodeName.equals(targetTag))
							found = true;
						else if (stopOnOthers)
							return false;
					}

//				    System.out.print("<"+nodeName+">");
				    startElement(nodeName);
					break;}

				case XMLEvent.END_ELEMENT: {
					 // stop parsing if we found the target node and its content is already processed.
					if (found)
						break parsingLoop;

					 // stop before leaving the targeted level range; used by nextStopOnOthers()
					if (_level == minLevel)
						return false;

					String nodeName = _reader.getLocalName();
				    endElement(nodeName);
//				    System.out.print("</"+nodeName+">");

					 // stop searching if the current level leaves the target range; used by back()
					if (_level < minLevel) {
						_reader.next();
						return false;
					}
					break;}

				case XMLEvent.CHARACTERS:
				case XMLEvent.COMMENT:
				case XMLEvent.SPACE:
				    characters();
//					System.out.print(_reader.getText());
					break;

				case XMLEvent.CDATA:
				    characters();
//					System.out.print(_reader.getText());
					break;

				case XMLEvent.PROCESSING_INSTRUCTION:
				    processingInstruction();
					break;
			}
		}

		if (found) {
			_node.setContent(_content);
			_content = "";
			return true;
		} else
			return false;
	}


	/**
	 * Find a children node by name.
	 * @param nodeName
	 * @return
	 * @throws XMLStreamException 
	 */
	public boolean find(final String nodeName) throws XMLStreamException
	{
		return firstChild(nodeName);
	}

	/**
	 * Find the n'th children node with tag nodeName.
	 * @param nodeName
	 * @param n
	 * @return
	 * @throws XMLStreamException 
	 */
	public boolean find(final String nodeName, int n) throws XMLStreamException
	{
		if (!firstChild(nodeName))
			return false;

		for(int i=1; i<n; ++i)
			if (!nextSkipOthers(nodeName))
				return false;

		return true;
	}

	/**
	 * Search matching child node identified by key name and an attribute value.
	 * @param nodeName
	 * @param attrName
	 * @param attrValue
	 * @return XMLNode
	 * @throws XMLStreamException 
	 */
	public boolean find(final String nodeName, final String attrName, final String attrValue) throws XMLStreamException
	{
		if (firstChild(nodeName)) {
    		if (_node.getName().equals(nodeName) && _node.get(attrName).equals(attrValue))
    			return true;

    		while(nextSkipOthers(nodeName))
    			if (_node.getName().equals(nodeName) && _node.get(attrName).equals(attrValue))
    				return true;
		}

		return false;
	}

	/**
	 * Search matching child node identified by key name and an attribute value.
	 * @param nodeName
	 * @param attrName
	 * @param attrValue
	 * @param n
	 * @return XMLNode
	 * @throws XMLStreamException 
	*/
	public boolean find(final String nodeName, final String attrName, final String attrValue, int n) throws XMLStreamException
	{
		if (firstChild(nodeName)) {
    		if (_node.getName().equals(nodeName) && _node.get(attrName).equals(attrValue))
			    if (n-- == 0)
			        return true;

    		while(nextSkipOthers(nodeName))
    			if (_node.getName().equals(nodeName) && _node.get(attrName).equals(attrValue))
    			    if (n-- == 0)
    			        return true;
		}

		return false;
	}

     /// XPath find function (const function)
	//TODO merge with XMLNode.findRelative()
    boolean findRelative(String path) throws XMLStreamException
    {
         // parse relative path
        while(path.length() > 0) {
            int idxSlash = path.indexOf('/');
            if (idxSlash == 0)
                return false;

            int l = idxSlash!=-1? idxSlash: path.length();
            String comp = path.substring(0, l);
            path = path.substring(l);

             // look for [n] and [@attrName="attrValue"] expressions in path components
            int idxBracket = comp.indexOf('[');
            l = idxBracket!=-1? idxBracket: comp.length();
            String childName = comp.substring(0, l);
            String attrName="", attrValue="";

            int n = 0;
            if (idxBracket != -1) {
                String expr = XMLHelper.unescape(comp.substring(idxBracket), '[', ']');

                try {
                    n = Integer.parseInt(expr);    // read index number
                } catch(NumberFormatException e) {
                    n = 0;
                }

                if (n != 0)
                    n = n - 1;  // convert into zero based index

                int idxAt = expr.indexOf('@');

                if (idxAt != -1) {
                    expr = expr.substring(idxAt+1);
                    int idxEqual = expr.indexOf('=');

                     // read attribute name and value
                    if (idxEqual != -1) {
                        attrName = XMLHelper.unescape(expr.substring(0, idxEqual));
                        attrValue = XMLHelper.unescape(expr.substring(idxEqual+1));
                    }
                }
            }

            boolean found;

            if (attrName.length() == 0)
                 // search n.th child node with specified name
                found = find(childName, n);
            else
                 // search n.th child node with specified name and matching attribute value
            	found = find(childName, attrName, attrValue, n);

            if (!found)
                return false;

            if (path.length()>0 && path.charAt(0)=='/')
                path = path.substring(1);
        }

        return true;
    }

}
