
//
// XML storage classes version 1.3
//
// XMLStreamWriter.java
//
// Copyright (c) 2004, 2005, 2006, 2007 Martin Fuchs <martin-fuchs@gmx.net>
//


/*

 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 * Redistributions of source code must retain the above copyright
	notice, this list of conditions and the following disclaimer.
 * Redistributions in binary form must reproduce the above copyright
	notice, this list of conditions and the following disclaimer in
	the documentation and/or other materials provided with the
	distribution.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.

*/

package de.mfuchs.xmlstorage.stream;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Writer;
import java.util.Map;
import java.util.Stack;

import de.mfuchs.xmlstorage.XMLHeader;
import de.mfuchs.xmlstorage.XMLHelper;
import de.mfuchs.xmlstorage.XMLNode;


/**
 * XMLStreamWriter is used to write XML streams in XMLStorage.
 */
public class XMLStreamWriter
{
     // PRETTY_FLAGS
    public static final int PRETTY_PLAIN    = 0;
    public static final int PRETTY_LINEFEED = 1;
    public static final int PRETTY_INDENT   = 2;


    public XMLStreamWriter(Writer out) throws IOException
	{
	    this(out, PRETTY_INDENT, new XMLHeader());
	}

    public XMLStreamWriter(final String path) throws IOException
    {
        this(path, PRETTY_INDENT, new XMLHeader());
    }

    public XMLStreamWriter(Writer out, int/*PRETTY_INDENT*/ pretty, final XMLHeader header) throws IOException
    {
        _out = out;
        _pretty = pretty;

        header.print(_out, false);
    }

    public XMLStreamWriter(final String path, int/*PRETTY_INDENT*/ pretty, final XMLHeader header) throws IOException
    {
         _out = new BufferedWriter(new FileWriter(path));
         _pretty = pretty;

        header.print(_out, false);
    }

    public void close() throws IOException
    {
        _out.write('\n');
        _out.close();
    }

     /// create node and move to it
    public void create(final String name) throws IOException
    {
        if (!_stack.empty()) {
            StackEntry last = (StackEntry) _stack.peek();

            if (last._state < PRE_CLOSED) {
                writeAttributes(last);
                closePre(last);
            }

            last._children = true;
        }

        StackEntry entry = new StackEntry(name);
        _stack.push(entry);

        writePre(entry);
    }

     /// go back to previous position
    public boolean back() throws IOException
    {
        if (!_stack.empty()) {
            writePost((StackEntry)_stack.peek());

            _stack.pop();

            return true;
        } else
            return false;
    }

     /// attribute setting
    public void put(String attrName, Object value)
    {
        if (!_stack.empty()) {
            XMLNode.AttributeMap attributes = ((StackEntry)_stack.peek())._attributes;

            if (value != null)
                attributes.put(attrName, value.toString());
            else
                attributes.remove(attrName);
        }
    }

     /// attribute setting
    public void put(String attrName, String value)
    {
        if (!_stack.empty()) {
            XMLNode.AttributeMap attributes = ((StackEntry)_stack.peek())._attributes;

            if (value != null)
                attributes.put(attrName, value);
            else
                attributes.remove(attrName);
        }
    }

    public void setContent(final String s)
    {
        setContent(s, false);
    }

    public void setContent(final String s, boolean cdata)
    {
        if (!_stack.empty())
            ((StackEntry)_stack.peek())._content = s!=null? XMLHelper.encodeXMLString(s, cdata): "";
    }

    // protected enum WRITESTATE
    protected static final int NOTHING     = 0;
    /*protected static final int PRE       = 1*/
    protected static final int ATTRIBUTES  = 2;
    protected static final int PRE_CLOSED  = 3;
    /*protected static final int CONTENT   = 4;*/
    protected static final int POST        = 5;

    protected Writer    _out;
    protected int/*PRETTY_FLAGS*/    _pretty;

    /**
     * StackEntry is used to hold state information when moving between nodes.
     */
    protected static class StackEntry {
        final String            _nodeName;
        XMLNode.AttributeMap    _attributes = new XMLNode.AttributeMap();
        String                  _content = "";
        int/*WRITESTATE*/       _state;
        boolean                 _children;

        StackEntry(String nodeName)
        {
            _nodeName = nodeName;
            _children = false;
            _state = NOTHING;
        }
    };

    protected Stack<StackEntry> _stack = new Stack<StackEntry>();

//  protected static final String s_empty_attr = "";

    protected void closePre(StackEntry entry) throws IOException
    {
        _out.write('>');

        entry._state = PRE_CLOSED;
    }

    protected void writePre(StackEntry entry) throws IOException
    {
        if (_pretty >= PRETTY_LINEFEED)
            _out.write('\n');

        if (_pretty == PRETTY_INDENT)
            for(int i=_stack.size(); --i>0; )
                _out.write(XMLNode.XML_INDENT_SPACE);

        _out.write("<" + XMLHelper.encodeXMLString(entry._nodeName));
        //entry._state = PRE;
    }

    protected void writeAttributes(StackEntry entry) throws IOException
    {
        for(Map.Entry<String,String> it: entry._attributes.entrySet())
            _out.write(" " + XMLHelper.encodeXMLString(it.getKey()) +
                       "=\"" + XMLHelper.encodeXMLString(it.getValue()) + "\"");

        entry._state = ATTRIBUTES;
    }

    protected void writePost(StackEntry entry) throws IOException
    {
        if (entry._state < ATTRIBUTES)
            writeAttributes(entry);

        if (entry._children || entry._content.length()>0) {
            if (entry._state < PRE_CLOSED)
                closePre(entry);

            _out.write(entry._content);
            //entry._state = CONTENT;

            if (_pretty>=PRETTY_LINEFEED && entry._content.length()==0)
                _out.write('\n');

            if (_pretty==PRETTY_INDENT && entry._content.length()==0)
                for(int i=_stack.size(); --i>0; )
                    _out.write(XMLNode.XML_INDENT_SPACE);

            _out.write("</" + XMLHelper.encodeXMLString(entry._nodeName) + ">");
        } else {
            _out.write("/>");
        }

        entry._state = POST;
    }
}
